# Performance: Scaling the Web with Apache #

This text has been transcribed from Igor Galić talk of the same title held at
the ApacheCon.eu 2012.

## Overview ##

To give you a feeling of what I'd like to talk about so you know whether you're
in the right session; This is a low-tech overview of what performance is, how to
measure it and how to improve it.

We'll start from the top, where it matters: The customer and work our way down
to the application, which for the most part we'll consider a black-box unless
it's reasonable to do otherwise.
We'll also talk about the application server, and the *System* itself, before I
finally invite you to discuss "architecture" with me.


## 1 Request ##

The most simple and important thing a Web-App needs to do is serve a single
request in a timely fashion. Often I'll get calls from angry or concerned
customers complaining how "the website is slow." The first thing I then do is
log onto the server, run top, find the server bored and dismiss my customer as
an idiot. If you're a freelancer this customer is paying your rent, instead of
dismissing them, try reliving their experience: Use a browser.

Most browsers these days come with built-in tools to measure how long each
request takes, and also provide us with information that lets us reason about
**why** it takes as long as it does.

A couple of years ago a colleague of mine, Theo Schlossnagle did a demo of
[YSlow! to Yfast in 45 minutes](http://omniti.com/seeds/yslow-to-yfast-in-45-minutes).
I'm going to attempt to compress that down to five.

### Caching Headers ###

Make sure that all reasonably cachable content is cachable downstream\*. For a month, for a year, for
ever. If a picture or a CSS file changes, give it another name.

(\* Downstream means either by "real" caches, like intermediary proxies, or "just"
by browsers)

### Etags ###

Make sure ETags are either configured to consistently deliver the same value, or
disable them. For files which fit into a single TCP packet you might want to
disable them altogether.

(This is an outstanding feature in Apache httpd: Implement file-size based
discrimination for [FileEtag](http://httpd.us.apache.org/docs/current/mod/core.html#fileetag).
Alternatively, one could use the mod_cache_socache
facilities to cache them. Patches welcome)


### Compress Content ###

There are a number of ways to compress content: One is to remove all whitespace
from CSS and JavaScript, thereby usually breaking all CSS and JavaScripts. (At
the very least it does a disservice to people who are trying to <del>steal</del>
learn from your code).

An easier way is to enable
[mod_deflate](http://httpd.apache.org/docs/current/mod/mod_deflate.html),
or Tomcat's [compress](http://tomcat.apache.org/tomcat-7.0-doc/config/http.html)
which compress
content on the fly. Or better yet, use [Traffic Server's gzip
plugin](http://trafficserver.apache.org/docs/trunk/admin/plugins/gzip/) which 
compresses content on the fly too, but then caches it.

You can also combine all CSS and JavaScript files to one file respectively,
this also reduces the number of round trips a browser needs to take to download
all content, and brings us to the next topic:

### Assets from Cookie-free Domain ###

You want to put your static content on a different domain. This comes with a
number of advantages:

Browsers restrict the numbers of concurrent connections they will open to a
single domain. Spreading static content to different domains "enables" them to
open more concurrent connections. Putting static content on a different domain
from dynamic also means that requests for static assets can happen without a
huge Cookie header. This again signals browsers, but more so downstream caches
that this content is cachable (for longer than the duration of the session).

Implementing these changes should notably reduce your Website's load times. But
this won't help much if your servers are in Germany, while your customers reside
in South America. This is where a CDN would come in real handy. How to build
and deploy one of those yourself is something I will talk in my "Introduction to
Apache Traffic Server"


## 500 Requests ##

Let's imagine your marketing department is running a campaign: They post a link
to Twitter/Facebook/Reddit/HN/. promoting a Web-App or a service. Your task is to
guarantee said service will actually sustain the load from all those (potential)
customers. Again, the easiest way to do this starts from customers perspective:

Clicking on the link on Twitter and recording the path you take with a tool such
a [tsung](http://www.process-one.net/en/tsung/) or
[JMeter](http://jmeter.apache.org/). For reasons I will not get more deeply into
that first link is often a Redirect. That's already two requests and you haven't
even started clicking through the intro or signing up for the App!

Once you're done recording you can replay your request (with different
parameters) 500 times in parallel, thousands times over.

This is the other reason I endorse tools like JMeter and tsung: Not only are
they multi-threaded, they also allow to use more than one machine, thereby
allowing you to do more than just benchmarking the CPU of the box you're running
the load test from, and instead allowing you to actually test the application.

While firing away with tsung, we need to take a close look at the test server's
health.

Recording `vmstat`, `mpstat`, `iostat`, etc, or aggregating all with `dstat`,
and visualizing those with either [fenxi](http://java.net/projects/fenxi) or
[vmstax](http://www.michenux.net/vmstax/).  Looking at a graphs rather than
thousands of numbers allows us to quickly reason about potential bottleneck.

Should "running `dstat`" not be enough to do so, I highly recommend Brandon
Gregg's [talks](http://www.slideshare.net/brendangregg/performance-use-method)
and [writings](http://dtrace.org/blogs/brendan/tag/usemethod/) on
[The USE Method](http://dtrace.org/blogs/brendan/2012/02/29/the-use-method/).

In general, though, a first assessment will usually tell us there's a bottleneck
with I/O.

## Grey Boxing ##

In order to identify exactly what the issue is we need to take a closer look:
But rather than second guessing by looking at the code, or filling it with debug
statements before we even know what's up, we orient ourselves on first-hand
evidence: Stack traces from the App itself running under load.

Most environments allow to the retrieval of stack traces to be as simple as
sending a `kill -ABRT`
([mod_passenger](http://www.modrails.com/documentation/Users%20guide%20Apache.html#debugging_frozen),
don't forget to set
[CoreDumpDirectory](http://httpd.apache.org/docs/current/mod/mpm_common.html#coredumpdirectory)
to a place where httpd has write permissions), or `kill -QUIT` (Tomcat, JBoss, ends
up in the servers' log). PHP has the very excellent [XDebug](http://www.xdebug.org).
If just looking at those traces isn't enough, then one can use [Samurai](http://yusuke.homeip.net/samurai/en/index.html),
(for Java) or [kcachegrind](http://kcachegrind.sourceforge.net/html/Home.html) (for PHP) to visualize
them.

Usually taking ~5 consecutive stack traces every 3s should suffice to make an
informed statement about an applications behaviour.

The function or method responsible for the (predominant) bottleneck will be on
top and will give us a hint as to which component this is:

A database, a cache, an external component. Verifying whether the bottleneck is
caused by that component or by the application's (wrong) use of it can be as
simple as (turning on a log and) looking at its logs. Should this not suffice,
we can, again, seek relief in the USE Method.

Fixing it thus means either turning a knob on such a component, e.g.: Setting
an Index to an often searched table in a Database, or fixing the code of our
application, i.e.: Rewriting said query.

This process can be repeated until all components that we control deliver a
satisfactory performance.

## Access Times ##

There is a reason that I/O issues are the first to become evident, and for a
long time in a debugging process remain the only obvious ones. That reason are
access times (or: latencies).

If, for convenience we imagine a single CPU cycle to take one second instead of
one nanosecond then access in and around the CPU happens within seconds (0 for
registers, 1s for L1, 10 for L2, 15 for L3 caches). Access to the main memory
already takes around five minutes, while disk access takes six months, if not a
year.

## Programming Environment, Application Server, OS, Machine ##

From many a programmer's perspective a program ends with their source code. Here
begins the abstract "system", strictly domain of the systems engineers.

One of the first choices when faced with improving an applications performance
is thus often to "upgrade" that system. After all the developers of these
"systems" are restlessly working on improving those components, and for most of
the time, they work *for* us. If a given web application is compatible with the
new release of the JDK, PHP Ruby, it is always preferable to update, simply
following the mantra:

> Infrastructure is an asset, code is liability.

An update from [JDK 6 to 7 can give a memory performance boost of
20%](http://hadoopstack.com/improving-hbase-read-performance/).
[PHP 5.3 to 5.4 saw as much as 50%](http://www.oracle.com/technetwork/articles/dsl/lerdorf-php54-1564639.html)
memory savings. Between 1.8 and 1.9 Ruby completely [replaced the
Interpreter](http://en.wikipedia.org/wiki/YARV), which brought a
[4-times performance boost](http://www.rubychan.de/share/yarv_speedups.html)
(according to synthetic benchmarks).

Next in line is the application server and besides running the latest and best
version of it, what determines its performance most of all are the various
settings for controlling and limiting process and thread spawning. Rather than
boring you about how to ideally tune those parameters for your application and
system, which their documentation already does very well, I simply want to
remind you of their *existence*:

The biggest problem with those parameters is that many users believe that they
come with sane, perhaps even *ideal* defaults. Instead, they just come with
*sample* values.

Apache httpd's mpm Configuration sets
[MaxRequestWorkers to 400](http://svn.apache.org/repos/asf/httpd/httpd/trunk/docs/conf/extra/httpd-mpm.conf.in).
We do this because we believe it's a good number for most systems found in the
wild these days. That is: It won't cause the system to blow up on startup. It
also illustrates how to configure the various mpms.

The Apache Tomcat project, through similar reasons, came to the conclusion that
200 would be a good default value for
[maxThreads and hence maxConnnections](http://tomcat.apache.org/tomcat-7.0-doc/config/http.html).
What happens now when we put an Apache httpd proxy in front of an Apache Tomcat,
once we reach 200 concurrent connections the Tomcat stops responding.
Similarly, once we reach 400 concurrent connections httpd, too stops
responding.
If you care for learning about defaults and their implications on a lower level,
I have pulled the `sysctl` values from two installations on identical machines
of the latest RHEL and Ubuntu and will do a hackathon/session on the subject
later this evening.

Rarely is it the CPU that becomes a real bottleneck and if it does it's often a
sign of either problem with the algorithm, in which case I have to refer you to
Donald E. Knuth's
[Art of Computer Programming](http://www-cs-faculty.stanford.edu/~knuth/taocp.html).
Or it could be a problem with the program's internal architecture.


## Architecture ##

But let's assume for a second that everything's working out fine and our
application server deliver 10K Requests/second. We put a cache, like Apache
Traffic Server in front and get some 200k maybe 500k on a good machine.

Unfortunately, for the load we're expecting (2-8m) this won't be enough:
Consequently we clone the setup on three more machines, put it behind a load
balancer we see a three-times decrease. Latencies are all over the place. What
went wrong is that we naïvely assumed this system was horizontally scalable.

I could share real-live horror stories about architecture gone fatally wrong,
but rather than reliving those (in front of an audience) I'd like to use this
space to promote one of my favourite NoSQL DBs:

A sadly often overlooked choice for both caching and storage. It can
effortlessly scale to multi-million, distributed nodes, is easily replicable
and when a node does go corrupt I as admin or application programmer could
hardly care less because restoring it can happen seamlessly from within the
application. I am, of course, talking about *CookieDB!*

If you want to learn more about
[Scaling Internet Architectures](http://scalableinternetarchitectures.com/), I
recommend you read Theo Schlossnagle's book of the same title, otherwise I
invite you to share your scaling "tricks" and if you have any Questions left:
Now's the time to ask them.

